"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ExtendedSchema = void 0;
const ValidationResults_1 = require("../models/ValidationResults");
const EMAIL_REGEX = new RegExp(/^(([^<>()[\]\\.,;:\s@"]+(\.[^<>()[\]\\.,;:\s@"]+)*)|(".+"))@((\[[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}])|(([a-zA-Z\-0-9]+\.)+[a-zA-Z]{2,}))$/);
const valueIsObject = (value) => {
    return typeof value === 'object' && !Array.isArray(value) && value !== null;
};
class ExtendedSchema {
    constructor(rules) {
        this.fields = {};
        this.rules = rules || {
            excess: 'clean',
        };
    }
    addField(key, type) {
        this.fields[key] = type;
    }
    addExtendedSchemaField(key, schema, callback, allowUndefined) {
        this.fields[key] = {
            typesToCheck: ['extSchema'],
            callback,
            allowUndefined,
            schema,
        };
    }
    // eslint-disable-next-line @typescript-eslint/explicit-module-boundary-types
    check(value) {
        var _a, _b;
        try {
            if (!valueIsObject(value)) {
                return (0, ValidationResults_1.WRONG_TYPE_INPUT)("The value isn't an object, but " + typeof value);
            }
            const valueKeys = Object.keys(value);
            const schemaKeys = Object.keys(this.fields);
            if (valueKeys.length > schemaKeys.length && ((_a = this.rules) === null || _a === void 0 ? void 0 : _a.excess) === 'forbid') {
                return (0, ValidationResults_1.EXCESS_KEYS_INPUT)('Schema does not allow excess keys. Excess keys: ' +
                    valueKeys.filter(k => !schemaKeys.includes(k)).join(', '));
            }
            const cleanValue = Object.assign({}, value);
            const encounteredKeys = new Set();
            for (const key in this.fields) {
                const field = this.fields[key];
                const valueToCheck = value[key];
                if (valueToCheck === undefined) {
                    if (!field.undefined || field.undefined === 'forbid') {
                        return (0, ValidationResults_1.WRONG_TYPE_INPUT)('Schema forbids undefined values. Key: ' + key);
                    }
                    if (field.callback && !field.callback(valueToCheck)) {
                        return (0, ValidationResults_1.CALLBACK_FAILED)('Callback failed. Key: ' + key);
                    }
                    encounteredKeys.add(key);
                    continue;
                }
                const type = typeof valueToCheck;
                let foundType = false;
                for (const typeToCheck of field.typesToCheck) {
                    if (type === typeToCheck) {
                        foundType = true;
                        // continue
                    }
                    if (typeToCheck === 'array' && Array.isArray(valueToCheck)) {
                        // because typeof [] === 'object'
                        foundType = true;
                        continue;
                    }
                    if (typeToCheck === 'null' && valueToCheck === null) {
                        foundType = true;
                        continue;
                    }
                    if (typeToCheck === 'extSchema' && valueIsObject(valueToCheck)) {
                        // extSchema does not support alternative types, so we can return early
                        if (!('schema' in field) || (field === null || field === void 0 ? void 0 : field.schema) === undefined) {
                            return (0, ValidationResults_1.INTERNAL_ERROR)('Schema is not defined for key: ' + key);
                        }
                        const result = field.schema.check(valueToCheck);
                        if (!result.success) {
                            return (0, ValidationResults_1.WRONG_TYPE_INPUT)('Nested schema failed. Key: ' + key);
                        }
                        foundType = true;
                        continue;
                    }
                    if (typeToCheck === 'object' && valueIsObject(valueToCheck)) {
                        foundType = true;
                    }
                }
                if (!foundType) {
                    return (0, ValidationResults_1.WRONG_TYPE_INPUT)('Type mismatch. Key: ' + key);
                }
                if (field.callback) {
                    const callbackResult = field.callback(valueToCheck);
                    if (!callbackResult) {
                        return (0, ValidationResults_1.CALLBACK_FAILED)('Callback failed. Key: ' + key);
                    }
                    else if (Array.isArray(callbackResult)) {
                        if (callbackResult.length !== 2 ||
                            typeof callbackResult[0] !== 'boolean' ||
                            typeof callbackResult[1] !== 'string') {
                            console.debug('Extended Schema encountered a bad callback result. Expected single boolean or [boolean, string], got: ', callbackResult);
                            return (0, ValidationResults_1.INTERNAL_ERROR)('Internal error');
                        }
                        const [success, message] = callbackResult;
                        if (!success) {
                            return (0, ValidationResults_1.CALLBACK_FAILED)(message);
                        }
                    }
                }
                encounteredKeys.add(key);
            }
            if (encounteredKeys.size < schemaKeys.length) {
                return (0, ValidationResults_1.MISSING_KEYS_INPUT)('Missing keys: ' + schemaKeys.filter(k => !encounteredKeys.has(k)).join(', '));
            }
            if (((_b = this.rules) === null || _b === void 0 ? void 0 : _b.excess) === 'clean') {
                for (const key in value) {
                    if (!encounteredKeys.has(key)) {
                        delete cleanValue[key];
                    }
                }
            }
            return (0, ValidationResults_1.VALID_INPUT)(cleanValue);
        }
        catch (error) {
            console.debug(error);
            return (0, ValidationResults_1.INTERNAL_ERROR)('Internal error');
        }
    }
    toJSON() {
        const result = {};
        for (const [key, value] of Object.entries(this.fields)) {
            const { typesToCheck, callback, undefined: canBeUndefined } = value;
            if (JSON.stringify(typesToCheck) === '["extSchema"]') {
                result[key] = value.schema.toJSON();
            }
            else {
                const { displayedAs } = value;
                const fieldKey = canBeUndefined === 'allow' ? `${key}?` : key;
                if (displayedAs) {
                    result[fieldKey] = displayedAs;
                }
                else {
                    let keyTypes = typesToCheck.join(' | ');
                    if (callback) {
                        keyTypes += ' (callback)';
                    }
                    result[fieldKey] = keyTypes;
                }
            }
        }
        return result;
    }
    length() {
        return Object.keys(this.fields).length;
    }
    /* SUGAR */
    addStringField(key, options) {
        this.addField(key, Object.assign(Object.assign({}, options), { typesToCheck: ['string'] }));
    }
    addNumberField(key, options) {
        this.addField(key, Object.assign(Object.assign({}, options), { typesToCheck: ['number'] }));
    }
    addNullField(key, options) {
        this.addField(key, Object.assign(Object.assign({}, options), { typesToCheck: ['null'] }));
    }
    addNullableField(key, otherTypes, options) {
        this.addField(key, Object.assign(Object.assign({}, options), { typesToCheck: [...otherTypes, 'null'] }));
    }
    addBooleanField(key, options) {
        this.addField(key, Object.assign(Object.assign({}, options), { typesToCheck: ['boolean'] }));
    }
    addArrayField(key, options) {
        this.addField(key, Object.assign(Object.assign({}, options), { typesToCheck: ['array'] }));
    }
    addEmailField(key, options) {
        this.addRegexField(key, EMAIL_REGEX, Object.assign(Object.assign({}, options), { displayedAs: (options === null || options === void 0 ? void 0 : options.displayedAs) || 'email' }));
    }
    addRegexField(key, regex, options) {
        this.addField(key, Object.assign(Object.assign({}, options), { typesToCheck: ['string'], callback: (value) => {
                if (!regex.test(value)) {
                    return false;
                }
                return (options === null || options === void 0 ? void 0 : options.callback) ? options.callback(value) : true;
            }, displayedAs: (options === null || options === void 0 ? void 0 : options.displayedAs) || 'REGEX string' }));
    }
    addFalsyField(key, options) {
        this.addBooleanField(key, Object.assign(Object.assign({}, options), { callback: (value) => {
                return !value;
            }, displayedAs: (options === null || options === void 0 ? void 0 : options.displayedAs) || 'false' }));
    }
    addTruthyField(key, options) {
        this.addBooleanField(key, Object.assign(Object.assign({}, options), { callback: (value) => {
                return !!value;
            }, displayedAs: (options === null || options === void 0 ? void 0 : options.displayedAs) || 'true' }));
    }
    addArrayOfElementsField(key, elementSchema, options) {
        this.addArrayField(key, Object.assign(Object.assign({}, options), { callback: (value) => {
                if (!Array.isArray(value)) {
                    return false;
                }
                for (const element of value) {
                    const result = elementSchema.check(element);
                    if (!result.success) {
                        return false;
                    }
                }
                return !(options === null || options === void 0 ? void 0 : options.callback) || options.callback(value);
            }, displayedAs: (options === null || options === void 0 ? void 0 : options.displayedAs) || 'array<' + JSON.stringify(elementSchema.toJSON()) + '>' }));
    }
}
exports.ExtendedSchema = ExtendedSchema;
